@@ -4,7 +4,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2020	PgPool Global Development Group
+ * Copyright (c) 2003-2023	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -451,7 +451,7 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 	else if (MAIN_REPLICA && query_context->is_multi_statement)
 	{
 		/*
-		 * If we are in native replication mode and we have multi statement query,
+		 * If we are in streaming replication mode and we have multi statement query,
 		 * we should send it to primary server only. Otherwise it is possible
 		 * to send a write query to standby servers because we only use the
 		 * first element of the multi statement query and don't care about the
@@ -484,7 +484,21 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 		/* Should be sent to both primary and standby? */
 		else if (dest == POOL_BOTH)
 		{
-			pool_setall_node_to_be_sent(query_context);
+			if (is_tx_started_by_multi_statement_query())
+			{
+				/*
+				 * If we are in an explicit transaction and the transaction
+				 * was started by a multi statement query, we should send
+				 * query to primary node only (which was supposed to be sent
+				 * to all nodes) until the transaction gets committed or
+				 * aborted.
+				 */
+				pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
+			}
+			else
+			{
+				pool_setall_node_to_be_sent(query_context);
+			}
 		}
 		else if (pool_is_writing_transaction() &&
 				 pool_config->disable_load_balance_on_write == DLBOW_ALWAYS)
@@ -527,34 +541,6 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 					 * Load balance if possible
 					 */
 
-					/*
-					 * As streaming replication delay is too much, if
-					 * prefer_lower_delay_standby is true then elect new
-					 * load balance node which is lowest delayed,
-					 * false then send to the primary.
-					 */
-					if (STREAM &&
-						pool_config->delay_threshold &&
-						bkinfo->standby_delay > pool_config->delay_threshold)
-					{
-						ereport(DEBUG1,
-								(errmsg("could not load balance because of too much replication delay"),
-								 errdetail("destination = %d for query= \"%s\"", dest, query)));
-
-						if (pool_config->prefer_lower_delay_standby)
-						{
-							int new_load_balancing_node = select_load_balancing_node();
-
-							session_context->load_balance_node_id = new_load_balancing_node;
-							session_context->query_context->load_balance_node_id = session_context->load_balance_node_id;
-							pool_set_node_to_be_sent(query_context, session_context->query_context->load_balance_node_id);
-						}
-						else
-						{
-							pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
-						}
-					}
-
 					/*
 					 * If system catalog is used in the SELECT, we prefer to
 					 * send to the primary. Example: SELECT * FROM pg_class
@@ -565,7 +551,7 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 					 * primary system catalog. Please note that this test must
 					 * be done *before* test using pool_has_temp_table.
 					 */
-					else if (pool_has_system_catalog(node))
+					if (pool_has_system_catalog(node))
 					{
 						ereport(DEBUG1,
 								(errmsg("could not load balance because systems catalogs are used"),
@@ -626,11 +612,44 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 					else
 					{
 						if (pool_config->statement_level_load_balance)
+						{
 							session_context->load_balance_node_id = select_load_balancing_node();
+							bkinfo = pool_get_node_info(session_context->load_balance_node_id);
+						}
 
-						session_context->query_context->load_balance_node_id = session_context->load_balance_node_id;
-						pool_set_node_to_be_sent(query_context,
-												 session_context->query_context->load_balance_node_id);
+						/*
+						 * As streaming replication delay is too much, if
+						 * prefer_lower_delay_standby is true then elect new
+						 * load balance node which is lowest delayed,
+						 * false then send to the primary.
+						 */
+						if (STREAM &&
+							pool_config->delay_threshold &&
+							bkinfo->standby_delay > pool_config->delay_threshold)
+						{
+							ereport(DEBUG1,
+									(errmsg("could not load balance because of too much replication delay"),
+									 errdetail("destination = %d for query= \"%s\"", dest, query)));
+
+							if (pool_config->prefer_lower_delay_standby)
+							{
+								int new_load_balancing_node = select_load_balancing_node();
+
+								session_context->load_balance_node_id = new_load_balancing_node;
+								session_context->query_context->load_balance_node_id = session_context->load_balance_node_id;
+								pool_set_node_to_be_sent(query_context, session_context->query_context->load_balance_node_id);
+							}
+							else
+							{
+								pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
+							}
+						}
+						else
+						{
+							session_context->query_context->load_balance_node_id = session_context->load_balance_node_id;
+							pool_set_node_to_be_sent(query_context,
+													 session_context->query_context->load_balance_node_id);
+						}
 					}
 				}
 				else
@@ -646,6 +665,10 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 			}
 		}
 	}
+	else if (REPLICATION && query_context->is_multi_statement)
+	{
+		pool_setall_node_to_be_sent(query_context);
+	}
 	else if (REPLICATION)
 	{
 		if (pool_config->load_balance_mode &&
@@ -756,24 +779,9 @@ pool_where_to_send(POOL_QUERY_CONTEXT * query_context, char *query, Node *node)
 	}
 
 	/*
-	 * EXECUTE?
+	 * DEALLOCATE or EXECUTE?
 	 */
-	if (IsA(node, ExecuteStmt))
-	{
-		POOL_SENT_MESSAGE *msg;
-
-		msg = pool_get_sent_message('Q', ((ExecuteStmt *) node)->name, POOL_SENT_MESSAGE_CREATED);
-		if (!msg)
-			msg = pool_get_sent_message('P', ((ExecuteStmt *) node)->name, POOL_SENT_MESSAGE_CREATED);
-		if (msg)
-			pool_copy_prep_where(msg->query_context->where_to_send,
-								 query_context->where_to_send);
-	}
-
-	/*
-	 * DEALLOCATE?
-	 */
-	else if (IsA(node, DeallocateStmt))
+	if (IsA(node, DeallocateStmt) || IsA(node, ExecuteStmt))
 	{
 		where_to_send_deallocate(query_context, node);
 	}
@@ -814,9 +822,9 @@ pool_send_and_wait(POOL_QUERY_CONTEXT * query_context,
 
 	/*
 	 * If the query is BEGIN READ WRITE or BEGIN ... SERIALIZABLE in
-	 * native replication mode, we send BEGIN to standbys instead.
-	 * original_query which is BEGIN READ WRITE is sent to primary.
-	 * rewritten_query which is BEGIN is sent to standbys.
+	 * streaming replication mode, we send BEGIN to standbys instead.
+	 * The original_query which is BEGIN READ WRITE is sent to primary.
+	 * The rewritten_query BEGIN is sent to standbys.
 	 */
 	if (pool_need_to_treat_as_if_default_transaction(query_context))
 	{
@@ -847,8 +855,8 @@ pool_send_and_wait(POOL_QUERY_CONTEXT * query_context,
 			continue;
 
 		/*
-		 * If in native replication mode, we do not send COMMIT/ABORT to
-		 * standbys if it's in I(idle) state.
+		 * If we are in streaming replication mode or logical replication mode,
+		 * we do not send COMMIT/ABORT to standbys if it's in I (idle) state.
 		 */
 		if (is_commit && MAIN_REPLICA && !IS_MAIN_NODE_ID(i) && TSTATE(backend, i) == 'I')
 		{
@@ -925,8 +933,15 @@ pool_send_and_wait(POOL_QUERY_CONTEXT * query_context,
 		 * Check if some error detected.  If so, emit log. This is useful when
 		 * invalid encoding error occurs. In this case, PostgreSQL does not
 		 * report what statement caused that error and make users confused.
+		 * Also set reset_query_error to true in ERROR case. This does
+		 * anything in normal query processing but when processing reset
+		 * queries, this is important because it might mean DISCARD ALL
+		 * command fails. If so, we need to discard the connection cache so
+		 * that any session object (i.e. named statement) does not remain in
+		 * the last session.
 		 */
-		per_node_error_log(backend, i, string, "pool_send_and_wait: Error or notice message from backend: ", true);
+		if (per_node_error_log(backend, i, string, "pool_send_and_wait: Error or notice message from backend: ", true) == 'E')
+			reset_query_error = true;
 	}
 
 	return POOL_CONTINUE;
@@ -967,9 +982,9 @@ pool_extended_send_and_wait(POOL_QUERY_CONTEXT * query_context,
 
 	/*
 	 * If the query is BEGIN READ WRITE or BEGIN ... SERIALIZABLE in
-	 * native replication mode, we send BEGIN to standbys instead.
-	 * original_query which is BEGIN READ WRITE is sent to primary.
-	 * rewritten_query which is BEGIN is sent to standbys.
+	 * streaming replication mode, we send BEGIN to standbys instead.
+	 * The original_query which is BEGIN READ WRITE is sent to primary.
+	 * The rewritten_query BEGIN is sent to standbys.
 	 */
 	if (pool_need_to_treat_as_if_default_transaction(query_context))
 	{
@@ -1360,8 +1375,19 @@ static POOL_DEST send_to_where(Node *node, char *query)
 			}
 			/* SAVEPOINT related commands are sent to both primary and standby */
 			else if (is_savepoint_query(node))
+			{
+				if (SL_MODE && is_tx_started_by_multi_statement_query())
+				{
+					/*
+					 * But in streaming replication mode, if a transaction was
+					 * started by a multi statement query, SAVEPOINT should be
+					 * sent to primary because the transaction was started on
+					 * primary only.
+					*/
+					return POOL_PRIMARY;
+				}
 				return POOL_BOTH;
-
+			}
 			/*
 			 * 2PC commands
 			 */
@@ -1520,26 +1546,55 @@ static POOL_DEST send_to_where(Node *node, char *query)
 	return POOL_PRIMARY;
 }
 
+/*
+ * Decide where to send given message.
+ * "node" must be a parse tree of either DEALLOCATE or EXECUTE.
+ */
 static
 void
 where_to_send_deallocate(POOL_QUERY_CONTEXT * query_context, Node *node)
 {
-	DeallocateStmt *d = (DeallocateStmt *) node;
+	DeallocateStmt *d = NULL;
+	ExecuteStmt *e = NULL;
+	char		*name;
 	POOL_SENT_MESSAGE *msg;
 
-	/* DEALLOCATE ALL? */
-	if (d->name == NULL)
+	if (IsA(node, DeallocateStmt))
 	{
-		pool_setall_node_to_be_sent(query_context);
+		d = (DeallocateStmt *) node;
+		name = d->name;
+	}
+	else if (IsA(node, ExecuteStmt))
+	{
+		e = (ExecuteStmt *) node;
+		name = e->name;
 	}
 	else
 	{
-		msg = pool_get_sent_message('Q', d->name, POOL_SENT_MESSAGE_CREATED);
+		ereport(ERROR,
+				(errmsg("invalid node type for where_to_send_deallocate")));
+		return;
+	}
+
+	/* DEALLOCATE ALL? */
+	if (d && (name == NULL))
+	{
+		/* send to all backend node */
+		pool_setall_node_to_be_sent(query_context);
+		return;
+	}
+
+	/* ordinary DEALLOCATE or EXECUTE */
+	else
+	{
+		/* check if message was created by SQL PREPARE */
+		msg = pool_get_sent_message('Q', name, POOL_SENT_MESSAGE_CREATED);
 		if (!msg)
-			msg = pool_get_sent_message('P', d->name, POOL_SENT_MESSAGE_CREATED);
+			/* message may be created by Parse message */
+			msg = pool_get_sent_message('P', name, POOL_SENT_MESSAGE_CREATED);
 		if (msg)
 		{
-			/* Inherit same map from PREPARE or PARSE */
+			/* Inherit same map from PREPARE or Parse */
 			pool_copy_prep_where(msg->query_context->where_to_send,
 								 query_context->where_to_send);
 
@@ -1547,8 +1602,33 @@ where_to_send_deallocate(POOL_QUERY_CONTEXT * query_context, Node *node)
 			query_context->load_balance_node_id = msg->query_context->load_balance_node_id;
 		}
 		else
-			/* prepared statement was not found */
-			pool_setall_node_to_be_sent(query_context);
+		{
+			/*
+			 * prepared statement was not found.
+			 * There are two cases when this could happen.
+			 * (1) mistakes by client. In this case backend will return ERROR
+			 * anyway.
+			 * (2) previous query was issued as multi-statement query. e.g.
+			 * SELECT 1\;PREPARE foo AS SELECT 1;
+			 * In this case pgpool does not know anything about the prepared
+			 * statement "foo".
+			 */
+			if (SL_MODE)
+			{
+				/*
+				 * In streaming replication or logical replication, sent to
+				 * primary node only.
+				 */
+				pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
+			}
+			else
+			{
+				/*
+				 * In other mode, sent to all node.
+				 */
+				pool_setall_node_to_be_sent(query_context);
+			}
+		}
 	}
 }
 
@@ -1732,10 +1812,10 @@ is_serializable(TransactionStmt *node)
 
 /*
  * If the query is BEGIN READ WRITE or
- * BEGIN ... SERIALIZABLE in native replication mode,
+ * BEGIN ... SERIALIZABLE in streaming replication mode,
  * we send BEGIN to standbys instead.
- * original_query which is BEGIN READ WRITE is sent to primary.
- * rewritten_query which is BEGIN is sent to standbys.
+ * The original_query which is BEGIN READ WRITE is sent to primary.
+ * The rewritten_query BEGIN is sent to standbys.
  */
 bool
 pool_need_to_treat_as_if_default_transaction(POOL_QUERY_CONTEXT * query_context)
