@@ -197,7 +197,7 @@ make_persistent_db_connection_noerror(
 		 * receives an ERROR, it stops processing and terminates, which is not
 		 * good. This is problematic especially with pcp_node_info, since it
 		 * calls db_node_role(), and db_node_role() calls this function. So if
-		 * the target PostgreSQL is down, EmitErrorReport() sends ERROR message
+		 * the target PostgreSQL is down, EmitErrorRepor() sends ERROR message
 		 * to pcp frontend and it stops (see process_pcp_response() in
 		 * src/libs/pcp/pcp.c. To fix this, just eliminate calling
 		 * EmitErrorReport(). This will suppress ERROR message but as you can
@@ -409,7 +409,8 @@ select_load_balancing_node(void)
 			pool_config->delay_threshold &&
 			pool_config->prefer_lower_delay_standby &&
 			(suggested_node_id != PRIMARY_NODE_ID) &&
-			(BACKEND_INFO(suggested_node_id).standby_delay > pool_config->delay_threshold))
+			(((BACKEND_INFO(suggested_node_id).standby_delay_by_time == false && BACKEND_INFO(suggested_node_id).standby_delay > pool_config->delay_threshold)) ||
+			 ((BACKEND_INFO(suggested_node_id).standby_delay_by_time && BACKEND_INFO(suggested_node_id).standby_delay > pool_config->delay_threshold_by_time * 1000000))))
 		{
 			ereport(DEBUG1,
 				(errmsg("selecting load balance node"),
